#define INCL_DOS
#define INCL_GPI
#define INCL_WIN
#define INCL_GREALL
//#define INCL_GRE_DEVMISC3
#include <os2.h>
#include <pmddi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define  _MEERROR_H_
#include <mmioos2.h>                   /* It is from MMPM toolkit           */
#include <dive.h>
//#include <fourcc.h>
#include "svgapmi.h"
#include "gradd.h"
#include "fslib.h"


// ---------------------------------------------------------------------------
// Shared global data
// ---------------------------------------------------------------------------

BOOL		fVideoStateSaved = FALSE,
		fVideoModeSwitched = FALSE;

APERTURE	aperture;
FBINFO		fbinfo;

PGDDMODEINFO	pgddModeInfo;
int		numVideoModes, defaultModeID;

HMODULE		hmodVMAN;
HMTX		hmtxDiveProtect;

FNVMIENTRY	*pfnVMIEntry;


// ---------------------------------------------------------------------------
// Service functions
// ---------------------------------------------------------------------------

int		unloadVMAN (void)
{
   DosFreeModule (hmodVMAN);

return 0;
}

int		loadVMAN (void)
{
   int rc;

   rc = DosLoadModule (NULL, 0, "VMAN", &hmodVMAN);
   if (rc) return rc; // unsuccess, possibly GRADD not installed

   rc = DosQueryProcAddr (hmodVMAN, 0, "VMIEntry", (PFN*)&pfnVMIEntry);

#ifdef FSDEBUG
   printf ("VMIEntry = %p (rc = %d)\n", pfnVMIEntry, rc);
   fflush (stdout);
#endif

return rc;
}


int		queryModeInfo (void)
{
   int rc, modeOperation;

   modeOperation = QUERYMODE_NUM_MODES;

   rc = pfnVMIEntry (0, VMI_CMD_QUERYMODES,
                     (PVOID)&modeOperation,
                     (PVOID)&numVideoModes);

   if (rc)
   {
      numVideoModes = 0;
      return rc;
   }

#ifdef FSDEBUG
   printf ("Available GRADD video modes: %d\n", numVideoModes);
   fflush (stdout);
#endif


   pgddModeInfo = (PGDDMODEINFO) malloc (numVideoModes * sizeof(*pgddModeInfo));

   if (pgddModeInfo)
   {
      int i;

      modeOperation = QUERYMODE_MODE_DATA;

      rc = pfnVMIEntry (0, VMI_CMD_QUERYMODES,
                        (PVOID)&modeOperation,
                        (PVOID)pgddModeInfo);
      if (!rc)
      {

#ifdef FSDEBUG
         for (i = 0; i < numVideoModes; i ++)
         {
            printf ("%dx%dx%d (%d)\n",
                    pgddModeInfo[i].ulHorizResolution,
                    pgddModeInfo[i].ulVertResolution,
                    pgddModeInfo[i].ulBpp,
                    pgddModeInfo[i].ulModeId);

            fflush(stdout); 
         }
#endif

      }
   return rc;
   }

return 1;
}

int	killPM (void)
{
   HAB  hab;
   HWND hwnd;
   HDC  hdc;

   int  rc;

   hab = WinQueryAnchorBlock (HWND_DESKTOP);
   hwnd = WinQueryDesktopWindow (hab, NULL);

   // Grab the handle to the Desktop DC
   hdc = WinOpenWindowDC (hwnd);

   // Call GreDeath
   rc = GreDeath (hdc);

return rc;
}

int	hardwarePointer (BOOL fState)
{
   HWSHOWPTRIN hwspi;
   int   rc;

   hwspi.ulLength = sizeof(hwspi);
   hwspi.fShow = fState;

   // hide hardware pointer
   rc = pfnVMIEntry (0, VMI_CMD_SHOWPTR, (PVOID)&hwspi, NULL);

return rc;
}

int	restorePM (void)
{
   HAB   hab;
   HWND  hwnd;
   HDC   hdc;
   int   rc;

   hab = WinQueryAnchorBlock (HWND_DESKTOP);

   // Get the window handle to the Desktop
   hwnd = WinQueryDesktopWindow (hab, NULL);

   // Grab the handle to the Desktop DC
   hdc = WinQueryWindowDC (hwnd);

   rc = GreResurrection (hdc, 0L, NULL);

   // Repaint the desktop
   WinInvalidateRect (hwnd, NULL, TRUE);

return rc;
}

BOOL	SetDisplayClut(PRGB2 prgb2Clut)
{
   HWPALETTEINFO hwpi;
   int rc;

   hwpi.ulLength = sizeof(hwpi);
   hwpi.fFlags = PALETTE_SET;
   hwpi.ulStartIndex = 0L;
   hwpi.ulNumEntries = 256L;
   hwpi.pRGBs = prgb2Clut;

   rc = pfnVMIEntry (0, VMI_CMD_PALETTE,
                     (PVOID)&hwpi,
                     NULL);

return (rc == 0);
}

BOOL	SwitchDiveMode (ULONG xRes, ULONG yRes, ULONG Bpp, PRGB2 prgb2Clut)
{
   int rc, i, switchModeID = -1;

   // dive protection
   if (hmtxDiveProtect)
      DosRequestMutexSem (hmtxDiveProtect, -1L);

   if (fVideoModeSwitched) // restore if already
   {
      RestoreDiveMode ();
   }

   for (i = 0; i < numVideoModes; i ++)
   {
      if ((pgddModeInfo[i].ulHorizResolution == xRes) &&
          (pgddModeInfo[i].ulVertResolution == yRes) &&
          (pgddModeInfo[i].ulBpp == Bpp))
      {
         switchModeID = pgddModeInfo[i].ulModeId;

#ifdef FSDEBUG
         printf ("Switching to mode: %dx%dx%d\n" \
                 "Refresh: %d\n" \
                 "Scanline size: %d\n" \
                 "Aperture size: %d\n" \
                 "VRAM physical address: %p",
                 xRes, yRes, Bpp,
                 pgddModeInfo[i].ulRefreshRate,
                 pgddModeInfo[i].ulScanLineSize,
                 pgddModeInfo[i].ulApertureSize,
                 pgddModeInfo[i].pbVRAMPhys);
         fflush (stdout);
#endif
         break;
      }
   }


   memset (&fbinfo, 0, sizeof(FBINFO));

   fbinfo.ulLength = sizeof(FBINFO);
   fbinfo.ulFlags = 0L;
   fbinfo.ulBPP = Bpp;
   fbinfo.ulXRes = xRes;
   fbinfo.ulYRes = yRes;
   fbinfo.ulScanLineBytes = pgddModeInfo[i].ulScanLineSize;
   fbinfo.ulENDIVEDrivers = 0L; // unknown to me ;-(

   switch (Bpp)
   {
      case 8L: fbinfo.fccColorEncoding = (ULONG)FOURCC_LUT8; break;

      case 16L: fbinfo.fccColorEncoding = (ULONG)FOURCC_R565; break;

      case 15L: fbinfo.fccColorEncoding = (ULONG)FOURCC_R555; break;

      case 24L: fbinfo.fccColorEncoding = (ULONG)FOURCC_RGB3; break;

      case 32L: fbinfo.fccColorEncoding = (ULONG)FOURCC_BGR4; break;
   }

   aperture.ulPhysAddr = (ULONG)pgddModeInfo[i].pbVRAMPhys;
   aperture.ulApertureSize = pgddModeInfo[i].ulApertureSize;
   aperture.ulScanLineSize = pgddModeInfo[i].ulScanLineSize;
   aperture.rctlScreen.yBottom = yRes - 1;
   aperture.rctlScreen.xRight = xRes - 1;
   aperture.rctlScreen.yTop = aperture.rctlScreen.xLeft = 0L;

   hardwarePointer (FALSE);
   killPM ();

   DosSleep (100L); // slip alittle

   rc = pfnVMIEntry (0, VMI_CMD_SETMODE,
                     (PVOID)&switchModeID,
                     NULL);

   if (!rc)
   {
      if ((Bpp == 8L) && prgb2Clut)
      {
         SetDisplayClut(prgb2Clut);
      }

      DiveFullScreenInit (&aperture, &fbinfo);

      fVideoModeSwitched = TRUE;
   } else restorePM(); // restore PM if error

   if (hmtxDiveProtect)
      DosReleaseMutexSem (hmtxDiveProtect);

return fVideoModeSwitched;
}

BOOL	RestoreDiveMode ()
{
   int rc;

   if(!fVideoModeSwitched) return FALSE;

   if (hmtxDiveProtect)
      DosRequestMutexSem (hmtxDiveProtect, -1L);

   rc = pfnVMIEntry (0, VMI_CMD_SETMODE,
                     (PVOID)&defaultModeID,
                     NULL);

   DosSleep (100L);

   DiveFullScreenTerm ();

   restorePM ();
   hardwarePointer (TRUE);

   fVideoModeSwitched = FALSE;

   if (hmtxDiveProtect)
      DosReleaseMutexSem (hmtxDiveProtect);

return TRUE;
}

void	 FsLibInit (HMTX diveProtect)
{
   if (!loadVMAN ())
   {
      INITPROCOUT ipo; // 10x to LightElf
      int rc;

      rc = pfnVMIEntry (0, VMI_CMD_INITPROC, NULL, (PVOID)&ipo);


      if (!queryModeInfo ())
      {
/*         DEVOPENSTRUC dop = {0L, (PSZ)"DISPLAY", NULL, 0L, 0L, 0L, 0L, 0L, 0L};
         HAB hab;
         HDC hdc;*/
         GDDMODEINFO defaultModeInfo;
         int i, bpp, width, height;

/*         width = WinQuerySysValue (HWND_DESKTOP, SV_CXSCREEN);
         height = WinQuerySysValue (HWND_DESKTOP, SV_CYSCREEN);

         hab = WinInitialize (0L);
         hdc = DevOpenDC (hab, OD_MEMORY, (PSZ)"*", 5L,
                          (PDEVOPENDATA)&dop, (HDC)NULL);

         DevQueryCaps (hdc,
                       CAPS_COLOR_BITCOUNT,
                       1L,
                       (PLONG)&bpp);

         DevCloseDC (hdc);
         WinTerminate (hab);
         
         for (i = 0; i < numVideoModes; i ++)
         {
            if ((pgddModeInfo[i].ulHorizResolution == width) &&
                (pgddModeInfo[i].ulVertResolution == height) &&
                (pgddModeInfo[i].ulBpp == bpp))
            {
               defaultModeID = pgddModeInfo[i].ulModeId;
               break;
            }
         }*/

         rc = pfnVMIEntry (0, VMI_CMD_QUERYCURRENTMODE,
                           NULL, (PVOID)&defaultModeInfo);

         for (i = 0; i < numVideoModes; i ++)
         {
            if (pgddModeInfo[i].ulModeId == defaultModeInfo.ulModeId)
            {
               width = pgddModeInfo[i].ulHorizResolution;
               height = pgddModeInfo[i].ulVertResolution;
               bpp = pgddModeInfo[i].ulBpp;
               break;
            }
         }

#ifdef FSDEBUG
         printf ("Current video mode: %dx%dx%d\n",
                 width, height, bpp);
         fflush (stdout);
#endif
         hmtxDiveProtect = diveProtect;
      }
   }
}