// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	System interface for sound.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_unix.c,v 1.5 1997/02/03 22:45:10 b1 Exp $";

#define INCL_DOS
#define INCL_DOSSEMAPHORES
#define INCL_DOSPROCESS
#define INCL_DOSMISC
#include <os2.h>
#include "windows.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <math.h>

#include <time.h>
#include <sys/types.h>

//#ifndef LINUX
//#include <sys/filio.h>
//#endif

#include <fcntl.h>
#include <io.h>
//#include <unistd.h>
//#include <sys/ioctl.h>

// Linux voxware output.
//#include <linux/soundcard.h>

// Timer stuff. Experimental.
#include <time.h>
#include <signal.h>

#include "z_zone.h"

#include "i_system.h"
#include "i_sound.h"
#include "m_argv.h"
#include "m_misc.h"
#include "w_wad.h"

#include "doomdef.h"

char WaveMsg[256];
void WriteDebug(char *);

#define VOICESSHIFT 5
#define VIRTUAL_VOICES (1<<VOICESSHIFT)

// Needed for calling the actual sound output.
#define SAMPLECOUNT    5120

// The actual lengths of all sound effects.
int lengths[NUMSFX];

extern int nosound;
extern int nomusic;

int sound_freq = 44100;
int sound_bits = 16;
int sound_buffer = 16384;
int sound_stereo = 1;

//
// this function converts raw 11khz, 8-bit data to a LPAUDIOWAVE that seal uses
// it is need cuz SEAL only loads samples from wavs
//added:11-01-98: now reads the frequency from the rawdata header.
//   dsdata points a 4 unsigned short header:
//    +0 : value 3 what does it mean?
//    +2 : sample rate, either 11025 or 22050.
//    +4 : number of samples, each sample is a single byte since it's 8bit
//    +6 : value 0

struct audio_wave {
unsigned sample_rate;
unsigned sample_size;
unsigned char *sampledata;
};

struct audio_wave *raw2SAMPLE(unsigned char *dsdata, int len, int priority)
{
    struct audio_wave *spl;
    unsigned char *p,*t;
    unsigned i;

    WriteDebug(".");

    spl = (struct audio_wave*)malloc(32);
    memset(spl, 0, sizeof(32));
    spl->sample_rate = *((short *)dsdata+1); // mostly 11025, but some at 22050.
    spl->sample_size = len-8;
    spl->sampledata = (void *)(dsdata+8);   // skip the 8bytes header

    return spl;
}

//using SEAL/2:
//
// This function loads the sound data from the WAD lump,
//  for single sound.
// This now returns a AUDIOWAVE*, not a void*
//
struct audio_wave * getsfx ( char*	       sfxname,
		  int*		len,
		  int		priority)
{
    unsigned char*	sfx;
    unsigned char*	paddedsfx;
    int 		i;
    int 		size;
    int 		paddedsize;
    char		name[20];
    int 		sfxlump;


    // Get the sound data from the WAD, allocate lump
    //	in zone memory.
    sprintf(name, "ds%s", sfxname);

    // Now, there is a severe problem with the
    //	sound handling, in it is not (yet/anymore)
    //	gamemode aware. That means, sounds from
    //	DOOM II will be requested even with DOOM
    //	shareware.
    // The sound list is wired into sounds.c,
    //	which sets the external variable.
    // I do not do runtime patches to that
    //	variable. Instead, we will use a
    //	default sound for replacement.
    if ( W_CheckNumForName(name) == -1 )
      sfxlump = W_GetNumForName("dspistol");
    else
      sfxlump = W_GetNumForName(name);

    size = W_LumpLength( sfxlump );

    // Debug.
    // fprintf( stderr, "." );
    //fprintf( stderr, " -loading  %s (lump %d, %d bytes)\n",
    //	     sfxname, sfxlump, size );
    //fflush( stderr );

    *len = size+SAMPLECOUNT;

    sfx = (unsigned char*)W_CacheLumpNum( sfxlump, PU_STATIC );

    return raw2SAMPLE(sfx, size, priority);
}





//
// This function adds a sound to the
//  list of currently active sounds,
//  which is maintained as a given number
//  (eight, usually) of internal channels.
// Returns a handle.
//
int addsfx ( int	   sfxid,
	     int	   volume,
	     int	   step,
	     int	   seperation )
{
// UNUSED
  return 0;
}


//
// SFX API
// Note: this was called by S_Init.
// However, whatever they did in the
// old DPMS based DOS version, this
// were simply dummies in the Linux
// version.
// See soundserver initdata().
//
void I_SetChannels()
{
  // UNUSED originaly for DMX and LINUX
}


void I_SetSfxVolume(int volume)
{
  if(nosound)
      return;

  snd_SfxVolume = volume;
//  Avolume(snd_SfxVolume*255/31,-1); // working !!!!!!!!!
audio_setvolume(volume);
}

// MUSIC API - dummy. Some code from DOS version.
void I_SetMusicVolume(int volume)
{
  if(nomusic)
     return;

midi_setvolume(volume);

  // Internal state variable.
  snd_MusicVolume = volume;
  // Now set volume on output device.
  // Whatever( snd_MusciVolume );
  // added 26-1-98
// working !!!!!!!!!!
}


//
// Retrieve the raw data lump index
//  for a given SFX name.
//
int I_GetSfxLumpNum(sfxinfo_t* sfx)
{
    char namebuf[9];
    sprintf(namebuf, "ds%s", sfx->name);
    return W_GetNumForName(namebuf);
}



//
// Starting a sound means adding it
//  to the current list of active sounds
//  in the internal channels.
// As the SFX info struct contains
//  e.g. a pointer to the raw data,
//  it is ignored.
// As our sound handling does not handle
//  priority, it is ignored.
// Pitching (that is, increased speed of playback)
//  is set, but currently not used by mixing.
//
int I_StartSound ( int		 id,
		   int		 vol,
		   int		 sep,
		   int		 pitch,
		   int		 priority )
{
  int i,voice;
  struct audio_wave *spl;

  if(nosound)
     return 0;

  // UNUSED
  priority = 0;

  spl = S_sfx[id].data;

  //play_sample(S_sfx[id].data,vol,sep,(pitch*1000)/128,0);

  voice = audio_play_sample(spl->sampledata, spl->sample_rate,
	                    spl->sample_size, vol, sep);


  // Returns a handle
  return (id<<VOICESSHIFT)+voice;
}

void I_StopSound (int handle)
{
  char msg[80];
  // You need the handle returned by StartSound.
  // Would be looping all channels,
  //  tracking down the handle,
  //  an setting the channel to zero.
  int voice=handle & (VIRTUAL_VOICES-1);

  if(nosound)
     return;

//  audio_stop_channel(voice);

//  mi13: debug stuff
//  sprintf(msg,"Stop voice %ld\n",voice);
//  WriteDebug(msg);

//  free(S_sfx[handle>>VOICESSHIFT].data);

//  if(voice_check(voice)==S_sfx[handle>>VOICESSHIFT].data)
//    deallocate_voice(voice);
}

int I_SoundIsPlaying(int handle)
{
  //BOOL stopped = FALSE;
  if(nosound)
  return FALSE;

  // return (!stopped);
  // BUG????? Hotfix by MadInt13h (seems to have a bug in SEAL sound engine)
  return TRUE;
}


//
// This function loops all active (internal) sound
//  channels, retrieves a given number of samples
//  from the raw sound data, modifies it according
//  to the current (internal) channel parameters,
//  mixes the per channel samples into the global
//  mixbuffer, clamping it to the allowed range,
//  and sets up everything for transferring the
//  contents of the mixbuffer to the (two)
//  hardware channels (left and right, that is).
//
//  allegro does this now
//
void I_UpdateSound( void )
{
}


//
// This would be used to write out the mixbuffer
//  during each game loop update.
// Updates sound buffer and audio device at runtime.
// It is called during Timer interrupt with SNDINTR.
// Mixing now done synchronous, and
//  only output be done asynchronous?
//

void I_SubmitSound( void )
{

}


void I_UpdateSoundParams( int	handle,
			  int	vol,
			  int	sep,
			  int	pitch)
{
  // I fail too see that this is used.
  // Would be using the handle to identify
  //  on which channel the sound might be active,
  //  and resetting the channel parameters.
  int voice=handle & (VIRTUAL_VOICES-1);
  int numsfx=handle>>VOICESSHIFT;

  if(nosound)
     return;

  audio_channel_parms(voice, vol, sep);

}


void I_ShutdownSound(void)
{
  // Wait till all pending sounds are finished.
  int i;

  if( nosound )
      return;

  audio_close();

//  sound_started = FALSE;
}

void I_StartupSound()
{
  int i,sfxcard,midicard;

  if(!nosound)
  {

  // Secure and configure sound device first.
  fprintf( stderr, "I_StartupSound: ");
  if (!audio_startup(sound_freq,sound_bits,sound_stereo,sound_buffer))
    fprintf(stderr,"audio startup failed!!!\n");
  else
    fprintf(stderr, " configured audio device\n" );

  // Initialize external data (all sounds) at start, keep static.
  fprintf( stderr, "I_InitSound: ");

  for (i=1 ; i<NUMSFX ; i++)
  {
    // Alias? Example is the chaingun sound linked to pistol.
    if (!S_sfx[i].link)
    {
      // Load data from WAD file.
      //added:11-01-98:note : this routine sets the second parm 'lengths'
      S_sfx[i].data = getsfx( S_sfx[i].name, &lengths[i],S_sfx[i].priority );
    }
    else
    {
      // Previously loaded already?
      S_sfx[i].data = S_sfx[i].link->data;
      lengths[i] = lengths[(S_sfx[i].link - S_sfx)/sizeof(sfxinfo_t)];
    }
  }
  fprintf( stderr, " pre-cached all sound data\n");

  // Finished initialization.
  fprintf(stderr, "I_InitSound: sound module ready\n");
  } else fprintf( stderr, "I_StartupSound: no sound device\n");
  //added:08-01-98:we use a similar startup/shutdown scheme as SEAL.
//  I_AddExitFunc(I_ShutdownSound);
//  sound_started = true;
}

void I_InitMusic(void)
{
  if(nomusic)
     return;

  fprintf( stderr, "I_InitMusic: ");
  if (midi_startup())
  {
     fprintf( stderr, "done\n");
     midi_setvolume(snd_MusicVolume);
  }
  else fprintf( stderr, "Error!!!");
}
void I_ShutdownMusic(void)	{ }

static int	looping=0;
static int	musicdies=-1;
void I_PauseSong(int handle)
{
  if(nomusic)
     return;

midi_pause(handle);
}
void I_ResumeSong(int handle)
{
  if(nomusic)
     return;

midi_resume(handle);
}
// Registers a song handle to song data.
//int I_RegisterSong(void *data);
int I_RegisterSong(void* data,int len)
{
int e, handle;
char *musicbuffer;

if(nomusic)
   return 0;

if(memcmp(data,"MUS",3)==0)
  {
    // convert mus to mid with a wanderfull function
    // thanks to S.Bacquet for the source of qmus2mid
    musicbuffer = (char*)malloc(256000); // well, something is wrong?
    if((e=qmus2mid((char *)data,musicbuffer,1,107,0,1,len))<0)
    {
      I_Error("Canot convert mus to mid, converterror :%d\n",e);
    }
    return midi_register(musicbuffer, e);
  }
  else
    I_Error("Music Lump is not MUS lump\n");

  return -1;

}
// Called by anything that wishes to start music.
//  plays a song, and when the song is done,
//  starts playing it again in an endless loop.
// Horrible thing to do, considering.
void
I_PlaySong
( int		handle,
  int		looping )
{
  if(nomusic)
     return;

midi_play(handle, looping);
midi_setvolume(snd_MusicVolume); // Hey, Pachucho!
}
// Stops a song over 3 seconds.
void I_StopSong(int handle)
{
  if(nomusic)
     return;

midi_stop(handle);
}
// See above (register), then think backwards
void I_UnRegisterSong(int handle)
{
  if(nomusic)
     return;

midi_deregister(handle);
}

