#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

typedef unsigned long DWORD;
typedef unsigned int WORD;
typedef unsigned char byte;
typedef unsigned char BYTE;

#define MIDI_BLOCK_SIZE 2048
#define MAX_BLOCKS 1

#define MEVT_F_SHORT        0x00000000L
#define MEVT_F_LONG         0x80000000L
#define MEVT_F_CALLBACK     0x40000000L

#define MEVT_EVENTTYPE(x)   ((BYTE)(((x)>>24)&0xFF))
#define MEVT_EVENTPARM(x)   ((DWORD)((x)&0x00FFFFFFL))

#define MEVT_SHORTMSG       ((BYTE)0x00)    /* parm = shortmsg for midiOutShortMsg */
#define MEVT_TEMPO          ((BYTE)0x01)    /* parm = new tempo in microsec/qn     */
#define MEVT_NOP            ((BYTE)0x02)    /* parm = unused; does nothing         */

/* 0x04-0x7F reserved */

#define MEVT_LONGMSG        ((BYTE)0x80)    /* parm = bytes to send verbatim       */
#define MEVT_COMMENT        ((BYTE)0x82)    /* parm = comment data                 */
#define MEVT_VERSION        ((BYTE)0x84)    /* parm = MIDISTRMBUFFVER struct       */


typedef struct
{
	DWORD	time;
	DWORD	ID;
	byte	data[3];
	byte	type;
}MidiEvent_t;

typedef struct
{
	DWORD	ID;          // identifier "MUS" 0x1A
	WORD	scoreLen;
	WORD	scoreStart;
	WORD	channels;	// count of primary channels
	WORD	sec_channels;	// count of secondary channels
	WORD	instrCnt;
	WORD	dummy;
}musheader_t;

/*typedef struct
{
	int			numevents;
	int			nextevent;
	MidiEvent_t	*midievents;
	MIDIHDR		header[2];
}songinfo_t;*/

byte		MidiControlers[10]={0, 0, 1, 7, 10, 11, 91, 93, 64, 67};
dboolean	started=false;

songinfo_t		*CurrentSong=NULL;
dboolean		loopsong=false;

byte XLateMUSControl(byte control)
{
	switch (control)
	{
	case 10:
		return(120);
	case 11:
		return(123);
	case 12:
		return(126);
	case 13:
		return(127);
	case 14:
		return(121);
	}
	fprintf(stderr,"Unknown control %d\n", control);
	return(0);
}

static int GetSongLength(byte *data)
{
	dboolean	done;
	int			events;
	musheader_t	*header;
	dboolean	time;

	header=(musheader_t *)data;
	data+=header->scoreStart;
	events=0;
	done=false;
	time=false;
	while (!done)
	{
		if (*data&0x80)
			time=true;
		switch((*(data++)>>4)&7)
		{
		case 1:
			if (*(data++)&0x80)
				data++;
			break;
		case 0:
		case 2:
		case 3:
			data++;
			break;
		case 4:
			data+=2;
			break;
		default:
			done=true;
			break;
		}
		events++;
		if (time)
		{
			while (*data&0x80)
				data++;
			data++;
			time=false;
		}
	}
	return(events+2);
}

static void Mus2Midi(byte *MusData, MidiEvent_t *MidiEvents)
{
	musheader_t	*header;
	byte		*score;
	int			spos;
	MidiEvent_t	*event;
	int			channel;
	int			etype;
	int			delta;
	int			finished;
	byte		channelvol[16];
	int			count;

	count=GetSongLength(MusData);
	header=(musheader_t *)MusData;
	if (header->ID!=0x1A53554D)//"MUS"<EOF>
		I_Error("Not a MUS file");
	score=(byte *)&MusData[header->scoreStart];
	event=MidiEvents;
	event->time=0;
	event->ID=0;
	event->type=MEVT_TEMPO;
	event->data[0]=0x00;
	event->data[1]=0x80;//not sure how to work this out, should be 140bpm
	event->data[2]=0x02;//but it's guessed so it sounds about right
	event++;

	event->time=0;
	event->ID=0;
	event->type=MEVT_TEMPO;
	event->data[0]=0x00;
	event->data[1]=0x80;//not sure how to work this out, should be 140bpm
	event->data[2]=0x02;//but it's guessed so it sounds about right
	event++;

	delta=0;
	spos=0;

	finished=false;
	while (true)
	{
		event->time=delta;
		delta=0;
		event->ID=0;
		etype=(score[spos]>>4)&7;
		event->type=MEVT_SHORTMSG;
		channel=score[spos]&15;
		if (channel==9)
			channel=15;
		else if (channel==15)
			channel=9;
		if (score[spos]&0x80)
			delta=-1;
		spos++;
		switch(etype)
		{
		case 0:
			event->data[0]=channel|0x80;
			event->data[1]=score[spos++];
			event->data[2]=channelvol[channel];
			break;
		case 1:
			event->data[0]=channel|0x90;
			event->data[1]=score[spos]&127;
			if (score[spos]&128)
				channelvol[channel]=score[++spos];
			spos++;
			event->data[2]=channelvol[channel];
			break;
		case 2:
			event->data[0]=channel|0xe0;
			event->data[1]=(score[spos]<<7)&0x7f;
			event->data[2]=score[spos++]>>7;
			break;
		case 3:
			event->data[0]=channel|0xb0;
			event->data[1]=XLateMUSControl(score[spos++]);
			event->data[2]=0;
			break;
		case 4:
			if (score[spos])
			{
				event->data[0]=channel|0xb0;
				event->data[1]=MidiControlers[score[spos++]];
				event->data[2]=score[spos++];
			}
			else
			{
				event->data[0]=channel|0xc0;
				spos++;
				event->data[1]=score[spos++];
				event->data[2]=64;
			}
			break;
		default:
			finished=true;
			break;
		}
		if (finished)
			break;

		event++;
		count--;
		if (count<0)
			I_Error("Overflow");
		if (delta==-1)
		{
			delta=0;
			while (score[spos]&128)
			{
				delta<<=7;
				delta+=score[spos++]&127;
			}
			delta+=score[spos++];
		}
	}
}
